// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <cassert>	// for assert

#include "Common.hpp"			 // for func_t
#include "Context.hpp"		 // for Context
#include "Debug.hpp"			 // for LOGD, LogSubsystem, LogSubsystem::DISP
#include "Emper.hpp"			 // for Emper::DEBUG
#include "Fiber.hpp"			 // for Fiber
#include "emper-common.h"	 // for workeraffinity_t

class Runtime;
class ContextManager;

class Dispatcher : public Logger<LogSubsystem::DISP> {
 protected:
	static thread_local const Fiber* currentFiber;

	Runtime& runtime;

	void dispatchLoopDoSleep();

	virtual void dispatchLoop() = 0;

	auto getDispatchLoop() -> func_t;

	// The dispatch() method could theoretically be made static in
	// non-debug builds.
	// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
	inline void dispatch(const Fiber* fiber) {
		LOGD("executing fiber " << fiber);
		currentFiber = fiber;

		if constexpr (emper::DEBUG) {
			Context::getCurrentContext()->setFiber(fiber);
		}

		fiber->run();
	}

	static inline auto isRunnable(Fiber* fiber) -> bool { return fiber->setRunnableFalse(); }

	static inline auto getAffinityBuffer(Fiber* fiber) -> workeraffinity_t* {
		return fiber->getAffinityBuffer();
	}

	static inline auto decreaseRefCount(Fiber* fiber) -> unsigned int {
		return fiber->doAtomicDecrRefCount();
	}

	static inline void recycle(const Fiber* fiber) { delete fiber; }

	void putRuntimeWorkerToSleep();

 public:
	Dispatcher(Runtime& runtime) : runtime(runtime) {}

	static auto getCurrentFiber() -> const Fiber& {
		const Fiber* fiber = getCurrentFiberPtr();
		return *fiber;
	}

	static auto getCurrentFiberPtr() -> const Fiber* {
		assert(currentFiber);
		return currentFiber;
	}

	static auto isDispatchedControlFlow() -> bool { return currentFiber != nullptr; }

	friend ContextManager;
};
