// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <atomic>				// for atomic_uint, atomic, __atomic_base, memory...
#include <cassert>			// for assert
#include <functional>		// for function
#include <iosfwd>				// for ostream
#include <type_traits>	// for remove_reference<>::type // IWYU pragma: keep
#include <utility>

#include "Common.hpp"			 // for ALIGN_TO_CACHE_LINE
#include "Debug.hpp"			 // for LogSubsystem, LogSubsystem::F, Logger
#include "emper-common.h"	 // for workeraffinity_t, UNUSED_ARG

#ifndef NDEBUG
#include <climits>	// for UINT_MAX
#endif

namespace adt {
template <typename T>
class MpscQueue;
}

#define FIBER_FUN_TEMPLATE_ARG void(void*)
#define FIBER_FUN0_TEMPLATE_ARG void(void)

class ALIGN_TO_CACHE_LINE Fiber : public Logger<LogSubsystem::F> {
 public:
	using fiber_fun_t = std::function<void(void*)>;
	using fiber_fun0_t = std::function<void()>;

	static const workeraffinity_t NOT_AFFINE = -1;

 private:
	const fiber_fun_t function;
	void* const arg;

	std::atomic<bool> runnable = {true};

	ALIGN_TO_CACHE_LINE std::atomic_uint referenceCounter = {1};

	workeraffinity_t* const affinity;

	Fiber* mpscNext = nullptr;

	/**
	 * A flag used to indicate where the fiber was from.
	 */
	unsigned int flag = 0;

	/**
	 * Dummy constructor. Used for example by the MpscQueue.
	 */
	Fiber() : Fiber([]() {}) {}

 protected:
	Fiber(fiber_fun_t function, void* arg, workeraffinity_t* affinity)
			: function(std::move(function)), arg(arg), affinity(affinity){};

	// cppcheck-suppress uninitMemberVar
	explicit Fiber(const fiber_fun0_t& function, workeraffinity_t* affinity)
			: Fiber([function](UNUSED_ARG void* arg) { function(); }, nullptr, affinity) {}

	Fiber(fiber_fun_t function, void* arg) : Fiber(std::move(function), arg, nullptr){};

	// cppcheck-suppress uninitMemberVar
	explicit Fiber(const fiber_fun0_t& function) : Fiber(function, nullptr) {}

	virtual ~Fiber() = default;

	virtual void run() const;

 private:
	inline void setMpscNext(Fiber* next) { mpscNext = next; }

	inline auto getMpscNext() -> Fiber* { return mpscNext; }

	[[nodiscard]] inline auto getAffinityBuffer() const -> workeraffinity_t* { return affinity; }

	inline auto setRunnableFalse() -> bool {
		bool res = runnable.load(std::memory_order_relaxed);
		if (!res) {
			// Fast path: 'runnable' was already set to false. No need
			// for an atomic operation.
			// Mesure: Impact of this "optimization". Is it really one?
			return res;
		}

		return runnable.exchange(false);
	}

	inline auto doAtomicIncrRefCount() -> unsigned int {
		assert(referenceCounter < UINT_MAX);
		return ++referenceCounter;
	}

	inline auto doAtomicDecrRefCount() -> unsigned int {
		assert(referenceCounter > 0);
		return --referenceCounter;
	}

	inline void setFlag(unsigned int flag) { this->flag = flag; }

	friend class adt::MpscQueue<Fiber>;
	friend class Scheduler;
	friend class Dispatcher;
	friend class LawsScheduler;

 public:
	[[nodiscard]] auto getFlag() const -> unsigned int { return flag; }

	[[nodiscard]] auto getAffinity() const -> workeraffinity_t {
		if (affinity == nullptr) {
			return NOT_AFFINE;
		}
		return *affinity;
	}

	void print() const;

	friend auto operator<<(std::ostream&, const Fiber&) -> std::ostream&;

	static inline auto from(fiber_fun_t function, void* arg) -> Fiber* {
		return new Fiber(std::move(function), arg);
	}

	static inline auto from(const fiber_fun0_t& function) -> Fiber* { return new Fiber(function); }

	static inline auto from(fiber_fun_t function, void* arg, workeraffinity_t* affinity) -> Fiber* {
		return new Fiber(std::move(function), arg, affinity);
	}

	static inline auto from(const fiber_fun0_t& function, workeraffinity_t* affinity) -> Fiber* {
		return new Fiber(function, affinity);
	}
};
