// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "Runtime.hpp"

#include <pthread.h>	// for pthread_t, pthread_attr_init

#include <cerrno>	 // for errno
// Non portable.
#include <sched.h>				// for cpu_set_t, CPU_SET, CPU_ZERO
#include <sys/sysinfo.h>	// for get_nprocs

#include <cstdlib>	// for rand, srand, abort
#include <cstring>
#include <memory>	 // for __shared_ptr_access, shared_ptr
#include <string>	 // for string

#include "Common.hpp"					 // for DIE_MSG_ERRNO, DIE, DIE_MSG
#include "ContextManager.hpp"	 // for ContextManager
#include "Debug.hpp"					 // for DBG, ABORT, LOGD, LOGE
#include "Emper.hpp"
#include "Fiber.hpp"								 // for Fiber
#include "RuntimeStrategy.hpp"			 // for RuntimeStrategy
#include "RuntimeStrategyStats.hpp"	 // for RuntimeStrategyStats
#include "emper-config.h"						 // IWYU pragma: keep
#include "lib/DebugUtil.hpp"
#include "strategies/ws/WsStrategy.hpp"	 // for WsStrategy, WsStrategy::INST...

#ifdef EMPER_LIBURCU
#include <urcu.h>	 // for rcu_register_thread
#endif

#ifndef EMPER_LOG_OFF
#include <syscall.h>	// for SYS_gettid
#include <unistd.h>		// for syscall

#include <ostream>	// for operator<<, basic_ostream<>:...
#endif

std::mutex Runtime::currentRuntimeMutex;
Runtime* Runtime::currentRuntime;
thread_local bool Runtime::workerThread = false;
thread_local unsigned int Runtime::seed;
thread_local workerid_t Runtime::workerId;
RuntimeStrategy& Runtime::DEFAULT_STRATEGY = WsStrategy::INSTANCE;

Runtime::Runtime(workerid_t workerCount, RuntimeStrategy& strategy, unsigned int seed)
		: workerCount(workerCount),
			workerLatch(workerCount),
			strategy(strategy),
			scheduler(strategy.getScheduler(*this)),
			dispatcher(strategy.getDispatcher(*this)),
			contextManager(*(new ContextManager(*this))),
			randomEngine(seed),
			atLeastOneWorkerIsSleeping(false),
			skipSleep(false) {
	threads = new pthread_t[workerCount];
	workerArgs = new struct WorkerArgs[workerCount];

	const int nprocs = get_nprocs();

	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		if (currentRuntime != nullptr) {
			DIE_MSG("There is already a runtime active");
		}
		currentRuntime = this;
	}

	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_attr_t attr;
		errno = pthread_attr_init(&attr);
		if (errno) DIE;

		// Start non portable.
		cpu_set_t cpuset;
		CPU_ZERO(&cpuset);
		CPU_SET(i % nprocs, &cpuset);

		errno = pthread_attr_setaffinity_np(&attr, sizeof(cpuset), &cpuset);
		if (errno) DIE_MSG_ERRNO("pthread_attr_setaffinity_np() failed");
		// End non portable.

		// Load the worker ID into the worker Argument array.
		workerArgs[i].id = i;
		// Load the worker PRNG seed into the Argument array.
		// This is not done by each thread individually because the randomEngine is
		// mutated and therefore produces data races.
		workerArgs[i].seed = uniformIntDistribution(randomEngine);

		auto thread_function = [](void* voidWorkerArgs) -> void* {
#ifdef EMPER_LIBURCU
			rcu_register_thread();
#endif

			auto* workerArgs = reinterpret_cast<struct WorkerArgs*>(voidWorkerArgs);
			return currentRuntime->workerLoop(workerArgs);
		};
		errno = pthread_create(&threads[i], &attr, thread_function, &workerArgs[i]);
		if (errno) DIE_MSG_ERRNO("pthread_create() failed");
	}

	if constexpr (emper::STATS) {
		int res = std::atexit(&printLastRuntimeStats);
		if (res) {
			DIE_MSG("could not register printStats() with at_exit()");
		}
	}

	const std::string stacktraceOnAbortsEnvVar("EMPER_STACKTRACE_ON_ABORTS");
	const char* envValueEnableStacktraces = std::getenv(stacktraceOnAbortsEnvVar.c_str());
	if (envValueEnableStacktraces && !std::strcmp(envValueEnableStacktraces, "true")) {
		DBG(stacktraceOnAbortsEnvVar << " set to 'true': Enabling stacktrace on abort");
		enableStacktraceOnAborts();
	}
}

Runtime::~Runtime() {
	DBG("Runtime " << this << " is terminating");
	for (workerid_t i = 0; i < workerCount; ++i) {
		DBG("Runtime " << this << " is cancelling worker " << unsigned(i));
		errno = pthread_cancel(threads[i]);
		if (errno) {
			DIE_MSG_ERRNO("pthread_cancel() failed");
		}
	}
	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		currentRuntime = nullptr;
	}
	DBG("Runtime " << this << " terminated");
}

auto Runtime::workerLoop(struct WorkerArgs* workerArgs) -> void* {
	workerThread = true;
	workerId = workerArgs->id;
	seed = workerArgs->seed;

	LOGD("Worker loop started by thread " << syscall(SYS_gettid));

	int oldType;
	errno = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldType);
	if (errno) {
		DIE_MSG_ERRNO("pthread_setcanceltype() failed");
	}

	for (const auto& f : newWorkerHooks) f();

	workerLatch.count_down_and_wait();

	contextManager.start();

	abort();
	return nullptr;
}

auto Runtime::nextFiber() -> Fiber* { return scheduler.nextFiber(); }

void Runtime::waitUntilFinished() {
	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_join(threads[i], nullptr);
	}
}

void Runtime::printStats() {
	auto runtimeStrategyStats = strategy.getStats();
	runtimeStrategyStats->print();
}

void Runtime::printLastRuntimeStats() {
	std::lock_guard<std::mutex> lock(currentRuntimeMutex);
	if (currentRuntime == nullptr) {
		LOGE("Could not print current runtime stats");
		return;
	}

	currentRuntime->printStats();
}

void Runtime::executeAndWait(std::function<void()> f) {
	if (inRuntime()) {
		ABORT("Ca not use executeAndWait() from within the Runtime");
	}

	std::mutex fiberFinished;
	fiberFinished.lock();

	Fiber* fiber = Fiber::from([&] {
		f();

		fiberFinished.unlock();
	});

	scheduleFromAnywhere(*fiber);

	fiberFinished.lock();
}
