// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <pthread.h>	// for pthread_t

#include <atomic>							 // for atomic, memory_order_relaxed
#include <cassert>						 // for assert
#include <condition_variable>	 // for condition_variable
#include <cstddef>						 // for size_t
#include <cstdint>						 // for intptr_t
#include <functional>					 // for function
#include <mutex>							 // for mutex, lock_guard, unique_lock
#include <random>
#include <thread>	 // for thread
#include <vector>	 // for vector

#include "Common.hpp"					 // for ALIGN_TO_CACHE_LINE
#include "Debug.hpp"					 // for LogSubsystem, LogSubsystem::RUNTI, Logger
#include "Scheduler.hpp"			 // for Scheduler
#include "emper-common.h"			 // for workerid_t
#include "lib/sync/Latch.hpp"	 // for Latch

class ContextManager;
class Dispatcher;
class Fiber;
class RuntimeStrategy;

enum WakeupMode {
	IF_SLEEPING_OBSERVED,
	ALWAYS,
};

class Runtime : public Logger<LogSubsystem::RUNTI> {
 private:
	static std::mutex currentRuntimeMutex;
	static Runtime* currentRuntime;

	static thread_local bool workerThread;
	static thread_local unsigned int seed;
	static thread_local workerid_t workerId;

	const workerid_t workerCount;

	std::vector<std::function<void(void)>> newWorkerHooks;

	Latch workerLatch;

	RuntimeStrategy& strategy;
	Scheduler& scheduler;
	Dispatcher& dispatcher;
	ContextManager& contextManager;
	pthread_t* threads;

	// Arguments passed to a new worker thread
	struct WorkerArgs {
		workerid_t id;
		unsigned int seed;
	};
	struct WorkerArgs* workerArgs;

	std::default_random_engine randomEngine;
	std::uniform_int_distribution<unsigned int> uniformIntDistribution;

	auto workerLoop(struct WorkerArgs* workerArgs) -> void*;

	ALIGN_TO_CACHE_LINE std::mutex workerSleepMutex;
	std::condition_variable workerSleepConditionVariable;
	ALIGN_TO_CACHE_LINE std::atomic<bool> atLeastOneWorkerIsSleeping;
	bool skipSleep;

	static RuntimeStrategy& DEFAULT_STRATEGY;

	static void printLastRuntimeStats();

 protected:
	void addNewWorkerHook(const std::function<void(void)>& hook) { newWorkerHooks.push_back(hook); };

	template <WakeupMode wakeupMode = WakeupMode::IF_SLEEPING_OBSERVED>
	inline void wakeupSleepingWorkers() {
		if constexpr (wakeupMode == WakeupMode::IF_SLEEPING_OBSERVED) {
			// If we observe no worker sleeping, then we do not try to
			// attempt to wakeup one. Note that this is racy, i.e., just
			// because we do not see any worker sleeping, does not mean that
			// one (or all but us, as a matter of a fact) is about to
			// sleep. However, in the worst case, this causes a worker to
			// sleep longer, until this worker (or another active worker)
			// schedules another fiber, when this method will be called
			// again. And then it is likely that the scheduling worker will
			// observe the sleeping worker.
			if (!atLeastOneWorkerIsSleeping.load(std::memory_order_relaxed)) {
				return;
			}
		}

		std::lock_guard<std::mutex> lk(workerSleepMutex);
		skipSleep = true;
		atLeastOneWorkerIsSleeping = false;
		workerSleepConditionVariable.notify_all();
	}

	void dispatcherLoopSleep() {
		if (skipSleep) {
			skipSleep = false;
			return;
		}

		std::unique_lock<std::mutex> lk(workerSleepMutex);
		// Check again if "skip sleep" has been set.
		if (skipSleep) return;

		atLeastOneWorkerIsSleeping = true;
		workerSleepConditionVariable.wait(lk);
	}

 public:
	Runtime() : Runtime(std::thread::hardware_concurrency()) {}

	Runtime(workerid_t workerCount) : Runtime(workerCount, DEFAULT_STRATEGY) {}

	Runtime(RuntimeStrategy& strategy) : Runtime(std::thread::hardware_concurrency(), strategy) {}

	Runtime(workerid_t workerCount, RuntimeStrategy& strategy,
					unsigned int seed = std::random_device()());

	~Runtime();

	inline void schedule(Fiber& fiber) {
		// Calling schedule() only works from within the EMPER runtime.
		assert(inRuntime());

		scheduler.schedule(fiber);
	}

	inline void scheduleFromAnywhere(Fiber& fiber) { scheduler.scheduleFromAnywhere(fiber); }

	auto nextFiber() -> Fiber*;

	// https://stackoverflow.com/a/3747462/194894
	static inline auto rand() -> int {
		seed = 214013 * seed + 2531011;
		auto shifted_seed = seed >> 16;
		return *reinterpret_cast<int*>(&shifted_seed) & 0x7FFF;
	}

	static inline auto getWorkerId() -> workerid_t { return workerId; }

	[[nodiscard]] inline auto getWorkerCount() const -> workerid_t { return workerCount; }

	static inline auto getRuntime() -> Runtime* { return currentRuntime; }

	inline auto getContextManager() -> ContextManager& { return contextManager; }

	inline auto getStrategy() -> RuntimeStrategy& { return strategy; }

	void waitUntilFinished();

	void printStats();

	static auto inRuntime() -> bool { return workerThread; }

	void executeAndWait(std::function<void()> f);

	friend ContextManager;
	friend Scheduler;
	friend Dispatcher;
	template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
	friend class MemoryManager;
};
