// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <mutex>
#include <queue>

#include "Blockable.hpp"
#include "CallerEnvironment.hpp"
#include "Context.hpp"

template <typename T>
class UnboundedBlockingMpscQueue : public Blockable<LogSubsystem::U_B_MPSC_Q> {
 private:
	std::atomic<Context*> blockedContext = nullptr;

	bool tPopped;
	T t;

	std::queue<T> mpscQueue;
	std::mutex queueMutex;

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void tryToWakeupBlockedContext() {
		Context* context = blockedContext.exchange(nullptr);
		if (!context) return;

		unblock<callerEnvironment>(context);
	}

	void tryToGetElement(const std::function<void(void)>& postRetrieve) {
		// tPopped indicates that 't' is a popped and usable value.
		// Therefore if we tryToGet a new 't' while 't' is available the current 't'
		// would be overridden and dropped.
		assert(!tPopped);
		std::lock_guard<std::mutex> lock(queueMutex);
		if (!mpscQueue.empty()) {
			t = mpscQueue.front();
			mpscQueue.pop();
			tPopped = true;
			postRetrieve();
		}
	}

 public:
	UnboundedBlockingMpscQueue(Runtime& runtime) : Blockable(runtime) {}

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void put(T t) {
		{
			std::lock_guard<std::mutex> lock(queueMutex);
			mpscQueue.push(t);
		}

		// Micro optimization, see if there is a blocked context
		// before performing the atomic exchange operation.
		if (blockedContext.load() != nullptr) {
			tryToWakeupBlockedContext<callerEnvironment>();
		}
	}

	void putFromAnywhere(T t) { put<CallerEnvironment::ANYWHERE>(t); }

	auto get(const std::function<void(void)>& postRetrieve) -> T {
		tPopped = false;
		tryToGetElement(postRetrieve);

		if (!tPopped) {
			Context* context = Context::getCurrentContext();
			block([this, context, postRetrieve] {
				tryToGetElement(postRetrieve);

				// N.B. It is important that we set blockedContext *after* we
				// called tryToGetElement(), as otherwise, multiple threads
				// may call tryToGetElement() which results in messages put in
				// the queue getting lost.
				blockedContext = context;

				if (tPopped) {
					tryToWakeupBlockedContext();
				}
			});
			if (!tPopped) {
				std::lock_guard<std::mutex> lock(queueMutex);
				// If 't' isn't already set, then mspcQueue.get() MUST
				// return an element. Note that a non-lineralizabe
				// queue may break this invariant.
				assert(!mpscQueue.empty());
				t = mpscQueue.front();
				mpscQueue.pop();
				postRetrieve();
			}
		}
		return t;
	}

	auto size() -> size_t {
		std::lock_guard<std::mutex> lock(queueMutex);
		return mpscQueue.size();
	}
};
