// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "LawsDispatcher.hpp"

#include <atomic>	 // for atomic, memory_order_relaxed

#include "Common.hpp"	 // for DIE_MSG
#include "Emper.hpp"
#include "Fiber.hpp"				 // for Fiber
#include "LawsStrategy.hpp"	 // for LawsStrategy, LawsStrategy::FiberSource
#include "Runtime.hpp"

void LawsDispatcher::dispatchLoop() {
	while (true) {
		Fiber* const fiber = runtime.nextFiber();
		if (!fiber) {
			dispatchLoopDoSleep();

			continue;
		}

		// The isRunnable() method uses an exchange on an atomic
		// boolean initialize to true in order to check if this fiber
		// is runnable.
		if (isRunnable(fiber)) {
			if constexpr (emper::STATS) {
				auto fiberSource = static_cast<LawsStrategy::FiberSource>(fiber->getFlag());
				switch (fiberSource) {
					case LawsStrategy::FiberSource::fromPriority:
						lawsStrategy.dispatchedFiberFromPriority.fetch_add(1, std::memory_order_relaxed);
						break;
					case LawsStrategy::FiberSource::fromLocal:
						lawsStrategy.dispatchedFiberFromLocal.fetch_add(1, std::memory_order_relaxed);
						break;
					case LawsStrategy::FiberSource::stolen:
						lawsStrategy.dispatchedFiberStolen.fetch_add(1, std::memory_order_relaxed);
						break;
					case LawsStrategy::FiberSource::anywhereQueue:
						lawsStrategy.dispatchedFiberFromAnywhere.fetch_add(1, std::memory_order_relaxed);
						break;
					default:
						DIE_MSG("Unknown fiber flag: " << flag);
						break;
				}
			}

			// The fiber was marked das runnable. Run it now.
			dispatch(fiber);
			// Update the affinity if one was set.
			workeraffinity_t* const affinity = getAffinityBuffer(fiber);
			if (affinity) *affinity = Runtime::getWorkerId();
		}

		if (decreaseRefCount(fiber) == 0) {
			recycle(fiber);
		}
	}
}
