// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "WsScheduler.hpp"

#include <atomic>
#include <ostream>

#include "Common.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "strategies/ws/WsStrategy.hpp"

class Fiber;

thread_local WsScheduler::WsQueue<WsScheduler::QUEUE_SIZE> WsScheduler::queue;

WsScheduler::WsScheduler(Runtime& runtime, WsStrategy& wsStrategy)
		: Scheduler(runtime), wsStrategy(wsStrategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	queues = new WsScheduler::WsQueue<QUEUE_SIZE>*[workerCount];

	auto newWorkerHook = [this]() { queues[Runtime::getWorkerId()] = &queue; };
	addNewWorkerHook(newWorkerHook);
}

void WsScheduler::schedule(Fiber& fiber) {
	LOGD("Scheduling fiber " << &fiber);

	bool pushed = queue.pushBottom(&fiber);
	if (unlikely(!pushed)) {
		// Work-stealing should not use an overflow queue
		// (EMPER_OVERFLOW_QUEUE), because of the extra overhead
		// required to check that queue for work, so we have to abort
		// here.
		ABORT("Could not push fiber " << &fiber << " into queue");
	}

	if constexpr (emper::STATS) {
		wsStrategy.scheduledFibers.fetch_add(1, std::memory_order_relaxed);
	}

	onNewWork();
}

auto WsScheduler::nextFiber() -> Fiber* {
	Fiber* fiber;
	bool poped = queue.popBottom(&fiber);

	if (likely(poped)) {
		if constexpr (emper::STATS) {
			wsStrategy.nextFiberFromLocal.fetch_add(1, std::memory_order_relaxed);
		}
		return fiber;
	}

	const workerid_t myWorkerId = Runtime::getWorkerId();
	const workerid_t workerCount = runtime.getWorkerCount();
	workerid_t startWorkerId = Runtime::rand() % workerCount;
	// TODO: See how reducing the loop bound affects things.
	for (workerid_t i = 0; i < workerCount; ++i) {
		workerid_t victim = (startWorkerId + i) % workerCount;

		// Don't steal from ourselves.
		if (unlikely(victim == myWorkerId)) continue;

		poped = queues[victim]->popTop(&fiber);
		if (poped) {
			if constexpr (emper::STATS) {
				wsStrategy.nextFiberStolen.fetch_add(1, std::memory_order_relaxed);
			}

			return fiber;
		}
	}

	// Try the "scheduled from anywhere" queue to get work as last resort.
	fiber = dequeFiberFromAnywhereQueue();
	return fiber;
}
